use std::num::NonZero;

pub type Node = NonZero<u32>;

#[derive(Debug, Clone, PartialEq, Eq)]
pub enum Tag {
    Undefined,
    Root,
    FunctionProto {
        /// Ident
        name: Node,
        /// ParameterList
        parameters: Option<Node>,
        return_type: Node,
    },
    ParameterList {
        /// [Parameter]
        parameters: Vec<Node>,
    },
    Parameter {
        /// Ident
        name: Node,
        /// TypeName
        ty: Node,
    },
    // TypeName: meta-tag of Ident | PrimitiveType | Pointer
    Pointer {
        /// TypeName
        pointee: Node,
    },
    PointerQualifier {
        constness: bool,
    },
    IntegralType(IntegralType),
    PrimitiveType(PrimitiveType),
    FunctionDecl {
        /// FunctionProto
        proto: Node,
        /// Block
        body: Node,
    },
    Ident {
        name: String,
    },
    IntegralConstant {
        bits: u64,
        ty: IntegralType,
    },
    FloatingConstant {
        bits: u64,
        ty: FloatingType,
    },
    Block {
        /// ReturnStmt | ExprStmt | VarDecl
        statements: Vec<Node>,
        trailing_expr: Option<Node>,
    },
    ReturnStmt {
        expr: Option<Node>,
    },
    ExprStmt {
        expr: Node,
    },
    VarDecl {
        let_or_var: LetOrVar,
        /// Ident
        name: Node,
        /// TypeName
        explicit_type: Option<Node>,
        assignment: Option<Node>,
    },
    DeclRef(Node),
    CallExpr {
        /// Ident | Expr
        lhs: Node,
        /// ArgumentList
        rhs: Option<Node>,
    },
    ArgumentList {
        /// [Argument]
        parameters: Vec<Node>,
    },
    Argument {
        /// Ident
        name: Option<Node>,
        /// expr
        expr: Node,
    },
    ExplicitCast {
        lhs: Node,
        /// TypeName
        typename: Node,
    },
    Deref {
        lhs: Node,
    },
    Ref {
        lhs: Node,
    },
    Not {
        lhs: Node,
    },
    Negate {
        lhs: Node,
    },
    Or {
        lhs: Node,
        rhs: Node,
    },
    And {
        lhs: Node,
        rhs: Node,
    },
    BitOr {
        lhs: Node,
        rhs: Node,
    },
    BitAnd {
        lhs: Node,
        rhs: Node,
    },
    BitXOr {
        lhs: Node,
        rhs: Node,
    },
    Eq {
        lhs: Node,
        rhs: Node,
    },
    NEq {
        lhs: Node,
        rhs: Node,
    },
    Lt {
        lhs: Node,
        rhs: Node,
    },
    Gt {
        lhs: Node,
        rhs: Node,
    },
    Le {
        lhs: Node,
        rhs: Node,
    },
    Ge {
        lhs: Node,
        rhs: Node,
    },
    Shl {
        lhs: Node,
        rhs: Node,
    },
    Shr {
        lhs: Node,
        rhs: Node,
    },
    Add {
        lhs: Node,
        rhs: Node,
    },
    Sub {
        lhs: Node,
        rhs: Node,
    },
    Mul {
        lhs: Node,
        rhs: Node,
    },
    Rem {
        lhs: Node,
        rhs: Node,
    },
    Div {
        lhs: Node,
        rhs: Node,
    },
    Assign {
        lhs: Node,
        rhs: Node,
    },
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum LetOrVar {
    Let,
    Var,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub struct IntegralType {
    pub signed: bool,
    pub bits: u16,
}

impl core::fmt::Display for IntegralType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}{}", if self.signed { "i" } else { "u" }, self.bits)
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum FloatingType {
    Binary32,
    Binary64,
}

impl core::fmt::Display for FloatingType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(match self {
            FloatingType::Binary32 => "binary32",
            FloatingType::Binary64 => "binary64",
        })
    }
}

impl IntegralType {
    pub fn u32() -> IntegralType {
        Self {
            signed: false,
            bits: 32,
        }
    }
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum Type {
    Any,
    Void,
    Bool,
    Integer(IntegralType),
    Floating(FloatingType),
    Pointer {
        constness: bool,
        pointee: Box<Self>,
    },
    Fn {
        parameter_types: Vec<Self>,
        return_type: Box<Self>,
    },
}

impl Type {
    pub fn void() -> Type {
        Self::Void
    }
    pub fn bool() -> Type {
        Self::Void
    }
    pub fn any() -> Type {
        Self::Void
    }
    pub fn into_ptr(self) -> Type {
        Self::Pointer {
            constness: false,
            pointee: Box::new(self),
        }
    }

    pub fn can_logical_and_or(&self) -> bool {
        match self {
            Type::Bool => true,
            _ => false,
        }
    }

    pub fn can_negate(&self) -> bool {
        match self {
            Type::Bool | Type::Integer(_) => true,
            _ => false,
        }
    }

    pub fn can_bitxor_and_or(&self) -> bool {
        match self {
            Type::Bool | Type::Integer(_) => true,
            _ => false,
        }
    }

    pub fn can_add_sub(&self) -> bool {
        match self {
            Type::Pointer { .. } | Type::Floating(_) | Type::Integer(_) => true,
            _ => false,
        }
    }
    pub fn can_shift(&self) -> bool {
        match self {
            Type::Integer(_) => true,
            _ => false,
        }
    }
    pub fn can_eq(&self) -> bool {
        match self {
            Type::Pointer { .. } | Type::Bool | Type::Floating(_) | Type::Integer(_) => true,
            _ => false,
        }
    }
    pub fn can_cmp(&self) -> bool {
        match self {
            Type::Pointer { .. } | Type::Floating(_) | Type::Integer(_) => true,
            _ => false,
        }
    }
    pub fn can_mul_div_rem(&self) -> bool {
        match self {
            Type::Floating(_) | Type::Integer(_) => true,
            _ => false,
        }
    }

    pub fn is_integer(&self) -> bool {
        match self {
            Type::Integer(_) => true,
            _ => false,
        }
    }
    pub fn remove_ptr(self) -> crate::parser::Result<Self> {
        match self {
            Type::Pointer { pointee, .. } => Ok(Box::into_inner(pointee)),
            _ => Err(crate::parser::Error::TriedToDerefNonPointer),
        }
    }
    pub fn align_of(&self) -> u32 {
        self.size_of()
    }
    pub fn size_of(&self) -> u32 {
        match self {
            Type::Any => 0,
            Type::Void => 0,
            Type::Bool => 1,
            Type::Integer(t) => t.bits.div_ceil(8) as u32,
            Type::Floating(t) => match t {
                FloatingType::Binary32 => 4,
                FloatingType::Binary64 => 8,
            },
            Type::Pointer { .. } => 8,
            Type::Fn { .. } => 8,
        }
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum PrimitiveType {
    FloatingType(FloatingType),
    IntegralType(Node),
    Bool,
    Void,
}

impl ToString for PrimitiveType {
    fn to_string(&self) -> String {
        match self {
            PrimitiveType::FloatingType(f) => f.to_string(),
            PrimitiveType::IntegralType(i) => i.to_string(),
            PrimitiveType::Bool => "bool".to_owned(),
            PrimitiveType::Void => "void".to_owned(),
        }
    }
}
